import datetime
import os
from abc import ABC, abstractmethod
from enum import Enum
from typing import List, Iterator

from fshof.exceptions import FshofException
from fshof.exist import file_exists, dir_exists
from fshof.expand import expand_path
from fshof.file_size import FileSize
from fshof.functional import FunctionalIterator


class FileLike(ABC):
    @property
    @abstractmethod
    def path(self) -> str:
        """
        Returns the path of the file-like object.

        :return: The path. Note that the path is maximally expanded (see the expand_path function).
        """
        raise NotImplementedError  # pragma: no cover

    def __str__(self) -> str:
        return repr(self)

    def __repr__(self) -> str:
        return str(self.path)


class File(FileLike):
    def __init__(self, path: str) -> None:
        if not file_exists(path):
            raise FshofException(f"There is no (regular) file at {path}")

        self._path = expand_path(path)
        assert os.path.isabs(self._path)

    @property
    def path(self) -> str:
        return self._path

    @property
    def dir(self) -> "Dir":
        """
        Return the directory containing this file.

        :return: A Dir object representing the directory.
        """
        return Dir(os.path.dirname(self._path))

    @property
    def extension(self) -> str:
        """
        The extension of this file.

        :return: The extension. If the file has no extension, an empty string will be
            returned. Otherwise the extension *without* the predecing dot will be returned
            (e.g. "txt", *not* ".txt")
        """
        _, ext = os.path.splitext(self._path)
        return ext[1:] if len(ext) > 0 else ext

    ext = extension

    @property
    def size(self) -> FileSize:
        """
        The size of this file.

        :return: A FileSize object representing the size of this file (in bytes).
        """
        return FileSize(os.path.getsize(self._path))

    @property
    def access_time(self) -> datetime.datetime:
        """
        The last access time of this file.

        :return: A datetime object representing the last access time.
        """
        atime = os.path.getatime(self._path)
        return datetime.datetime.fromtimestamp(atime)

    atime = access_time

    @property
    def mod_time(self) -> datetime.datetime:
        """
        The last modification time of this file.

        :return: A datetime object representing the last modification time.
        """
        mtime = os.path.getmtime(self._path)
        return datetime.datetime.fromtimestamp(mtime)

    mtime = mod_time

    def text_file(self, encoding: str = "utf-8") -> "TextFile":
        """
        Obtain a TextFile object for this file.

        You are therefore responsible to ensure that the underlying file is a valid
        textfile (since this is very expensive to ensure automatically). This function
        will always succeed, even if the underlying file is not a valid text file.
        However, when calling functions on such a TextFile object, errors will occur.

        :param encoding: The encoding to use.
        :type encoding: str
        :return: The obtained TextFile object.
        """
        return TextFile(self._path, encoding)

    t = text_file


class TextFile(File):
    def __init__(self, path: str, encoding: str = "utf-8"):
        super().__init__(path)

        self.encoding = encoding

    @property
    def content(self) -> str:
        """
        The content of the file.

        :return: The content.
        """
        with open(str(self.path), "r", encoding=self.encoding) as regular_file:
            return regular_file.read()

    @property
    def lines(self) -> List[str]:
        """
        The lines of this file.

        :return: A list of lines.
        """
        with open(str(self.path), "r", encoding=self.encoding) as regular_file:
            return regular_file.readlines()

    @property
    def words(self) -> List[str]:
        """
        The words of this file. It is assumed that words are separated by whitespace.

        :return: A list of words.
        """
        return self.content.split()

    @property
    def char_count(self) -> int:
        """
        The number of characters of this file.

        :return: The number of characters.
        """
        return len(self.content)

    cc = char_count

    @property
    def word_count(self) -> int:
        """
        The number of words of this file.

        :return: The number of words.
        """
        return len(self.words)

    wc = word_count

    @property
    def line_count(self) -> int:
        """
        The number of lines of this file.

        :return: The number of lines.
        """
        return len(self.lines)

    lc = line_count


class FileTreeWalkIteratorKind(Enum):
    BOTH = 0
    REGULAR_FILES_ONLY = 1
    DIRS_ONLY = 2


class FileTreeWalkIterator(Iterator):
    def __init__(self, path: str, kind: FileTreeWalkIteratorKind) -> None:
        self.it = os.walk(path)
        self.path = path

        self.kind = kind

        self.sub_dir_path, _, file_names = next(self.it)
        self.sub_dir_processed = False
        self.file_names = sorted(file_names)

    def __next__(self) -> FileLike:
        if not self.sub_dir_processed and (
            self.kind == FileTreeWalkIteratorKind.BOTH
            or self.kind == FileTreeWalkIteratorKind.DIRS_ONLY
        ):
            self.sub_dir_processed = True
            return Dir(self.sub_dir_path)

        if len(self.file_names) != 0 and (
            self.kind == FileTreeWalkIteratorKind.BOTH
            or self.kind == FileTreeWalkIteratorKind.REGULAR_FILES_ONLY
        ):
            file_name, *self.file_names = self.file_names
            return File(os.path.join(self.sub_dir_path, file_name))

        try:
            self.sub_dir_path, _, file_names = next(self.it)
            self.file_names = sorted(file_names)
            self.sub_dir_processed = False
            return next(self)
        except StopIteration:
            raise StopIteration


class Dir(FileLike):
    def __init__(self, dir_path: str) -> None:
        if not dir_exists(dir_path):
            raise FshofException(f"There is no directory at {dir_path}")

        self.normal_path = expand_path(dir_path)
        assert os.path.isabs(self.normal_path)

    def file(self, file_name: str) -> "File":
        """
        A file
        :param file_name:
        :return:
        """
        return File(os.path.join(self.normal_path, file_name))

    def dir(self, subdir_name: str) -> "Dir":
        return Dir(os.path.join(self.normal_path, subdir_name))

    @property
    def path(self) -> str:
        return self.normal_path

    @property
    def file_likes(self) -> FunctionalIterator[FileLike]:
        """
        An iterator of file-like objects present in this directory and all subdirectories.

        :return: The iterator.
        """
        return FunctionalIterator(
            FileTreeWalkIterator(self.normal_path, FileTreeWalkIteratorKind.BOTH)
        )

    @property
    def files(self) -> "FileIterator":
        """
        An iterator of (regular) files present in this directory and all subdirectories.

        :return: The iterator.
        """
        return FileIterator(
            FileTreeWalkIterator(
                self.normal_path, FileTreeWalkIteratorKind.REGULAR_FILES_ONLY
            )
        )

    @property
    def dirs(self) -> FunctionalIterator["Dir"]:
        """
        An iterator of all directories present in this directory and all subdirectories.

        :return: The iterator.
        """
        return FunctionalIterator(
            FileTreeWalkIterator(self.normal_path, FileTreeWalkIteratorKind.DIRS_ONLY)
        )


class FileIterator(FunctionalIterator["File"]):
    def filter_extension(self, extension: str) -> "FileIterator":
        return FileIterator(self.filter(lambda file: file.extension == extension))

    filter_ext = filter_extension

    def map_char_count(self) -> FunctionalIterator[int]:
        return self.map(lambda f: f.text_file().char_count)

    map_cc = map_char_count

    def map_word_count(self) -> FunctionalIterator[int]:
        return self.map(lambda f: f.text_file().word_count)

    map_wc = map_word_count

    def map_line_count(self) -> FunctionalIterator[int]:
        return self.map(lambda f: f.text_file().line_count)

    map_lc = map_line_count
